/*
 * coverart.c
 * 
 * Copyright (C) 2010 Sinew Software Systems
 * Authers: Vinod Kumar Dahiya
 *          Hemant Kumar
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The Rhythmbox authors hereby grant permission for non-GPL compatible
 * GStreamer plugins to be used and distributed together with GStreamer
 * and Rhythmbox. This permission is above and beyond the permissions granted
 * by the GPL license by which Rhythmbox is covered. If you modify this code
 * you may extend this exception to your version of the code, but you are not
 * obligated to do so. If you do not wish to do so, delete this exception
 * statement from your version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 */

#include <gio/gio.h>
#include <gst/gst.h>
#include <glib/gi18n-lib.h>

#include "coverart.h"
#include "rhythmarty.h"

static void 
rb_arty_cover_progress_step(RBArtyPlugin *plugin);
static void 
rb_arty_cover_progress_finish(RBArtyPlugin *plugin);

typedef struct _ArtyEmbeddedCover
{
	GstElement *pipeline;
	GstBus* bus;
	GstElement *src;
	GstElement *decodebin;
	GstElement *sink;
	GstPad *sinkpad;
	gchar *album_name;
	RBArtyPlugin *plugin;
}ArtyEmbeddedCover;

static void
rb_arty_save_coverart(RBArtyPlugin *plugin, const gchar *album_name,
                      GdkPixbuf *pixbuf)
{
	gchar *escaped = g_uri_escape_string(album_name,NULL,FALSE);
	gchar *filename=g_strconcat(plugin->priv->cache_dir,"/",
	                            escaped,".jpg",NULL);

	gdk_pixbuf_save (pixbuf, filename, "jpeg", NULL, NULL);

	g_free(escaped);
	g_free(filename);
}

void arty_cover_free(ArtyEmbeddedCover *cover)
{
	gst_element_unlink(cover->src,cover->decodebin);
	gst_bin_remove(GST_BIN(cover->pipeline),cover->src);
	g_free(cover->album_name);
}

void tag_callback(GstBus *bus,GstMessage *message, ArtyEmbeddedCover *cover)
{
	//g_print("got tags\n");
	GstTagList *list;
	gst_message_parse_tag (message,&list);
	
	const gchar *tags[]={
		GST_TAG_IMAGE,
		GST_TAG_PREVIEW_IMAGE
	};
	
	int i;
	for(i=0;i<G_N_ELEMENTS(tags);i++)
	{
		if(!gst_tag_list_get_tag_size(list,tags[i]))
			continue;
					
		GstBuffer *buffer;
		if(gst_tag_list_get_buffer(list,tags[i],&buffer))
		{
			GdkPixbufLoader *loader= gdk_pixbuf_loader_new ();
			gdk_pixbuf_loader_write (loader,buffer->data,buffer->size,NULL);
			gdk_pixbuf_loader_close (loader,NULL);
			
			GdkPixbuf *pixbuf=NULL;
			pixbuf = 
				gdk_pixbuf_scale_simple (gdk_pixbuf_loader_get_pixbuf(loader),
				                         200,200,GDK_INTERP_BILINEAR);
			g_object_unref (loader);
			
			if(pixbuf)
			{
				//g_print("From embedded : %s\n",cover->album_name);
				rb_arty_save_coverart (cover->plugin,
				                         cover->album_name,pixbuf);
				
				g_object_unref(pixbuf);
				gst_buffer_unref(buffer);
				break;
			}
			
			gst_buffer_unref(buffer);
		}
	}
	
	gst_tag_list_free(list);
}

void state_changed_callback(GstBus *bus,GstMessage *message, ArtyEmbeddedCover *cover)
{	
	if(GST_ELEMENT(message->src)!=cover->pipeline)
		return;
		
	//g_print("state changed\n");
	
	GstState old,new,pending;
	gst_message_parse_state_changed (message,&old,&new,&pending);
	
	if(new == GST_STATE_PAUSED && pending == GST_STATE_VOID_PENDING)
	{
		//g_print("setting null\n");
		
		cover->plugin->priv->cover_fetcher->idle_running = FALSE;
		
		gst_element_set_state (cover->pipeline,GST_STATE_NULL);
		arty_cover_free(cover);

		rb_arty_cover_fetcher_fetch_next(cover->plugin->priv->cover_fetcher);
	}
}

void error_callback(GstBus *bus,GstMessage *message, ArtyEmbeddedCover *cover)
{
	g_print("RhythmArty gst error\n");

	cover->plugin->priv->cover_fetcher->idle_running = FALSE;
	
	gst_element_set_state (cover->pipeline,GST_STATE_NULL);
	arty_cover_free(cover);

	rb_arty_cover_fetcher_fetch_next(cover->plugin->priv->cover_fetcher);
}

void decoded_pad_callback(GstElement *decodebin,GstPad *pad,gboolean islast,ArtyEmbeddedCover *cover)
{
	if(gst_pad_is_linked(cover->sinkpad))
		return;
		
	//g_print("decoded linking ..\n");
	gst_pad_link(pad,cover->sinkpad);
}

void init_gstreamer_elements(RBArtyPlugin *plugin)
{
	if(plugin->priv->cover_fetcher->cover)
		return;

	//g_print("initilising gstreamer ...\n");
	
	plugin->priv->cover_fetcher->cover = g_new0(ArtyEmbeddedCover,1);
	
	ArtyEmbeddedCover *gcover=plugin->priv->cover_fetcher->cover;
	gcover->plugin = plugin;
	gcover->album_name = NULL;
	gcover->pipeline = gst_pipeline_new ("ArtFetcher");
	gcover->bus = gst_pipeline_get_bus (GST_PIPELINE(gcover->pipeline));
	
	gst_bus_add_signal_watch (gcover->bus);
	
	g_signal_connect (gcover->bus,"message::tag",G_CALLBACK(tag_callback),gcover);
	g_signal_connect (gcover->bus,"message::state-changed",G_CALLBACK(state_changed_callback),gcover);
	g_signal_connect (gcover->bus,"message::error",G_CALLBACK(error_callback),gcover);
	
	gcover->src = NULL;
	gcover->decodebin = gst_element_factory_make("decodebin2","ArtFetcher-decoder");
	gcover->sink = gst_element_factory_make("fakesink","ArtFetcher-sink");
	
	gst_bin_add_many(GST_BIN(gcover->pipeline),gcover->decodebin,gcover->sink,NULL);

	gcover->sinkpad = gst_element_get_static_pad (gcover->sink,"sink");
	g_signal_connect (gcover->decodebin,"new-decoded-pad",G_CALLBACK(decoded_pad_callback),gcover);
}

void rb_arty_get_embedded_cover(RBArtyPlugin *plugin,
                                const gchar *album_name,
                                const gchar *uri)
{	
	init_gstreamer_elements(plugin);

	ArtyEmbeddedCover *gcover=plugin->priv->cover_fetcher->cover;
	
	gcover->album_name = g_strdup (album_name);
	
	gcover->src = gst_element_make_from_uri(GST_URI_SRC,uri,"ArtFetcher-src");
	
	gst_bin_add(GST_BIN(gcover->pipeline),gcover->src);
	gst_element_link(gcover->src,gcover->decodebin);
	
	gst_element_set_state (gcover->pipeline,GST_STATE_PAUSED);	
}

GdkPixbuf *rb_arty_get_album_art_from_location(const gchar *location)
{
	const char* cover_names[]={
		"cover.jpg", "album.jpg", "albumart.jpg", ".folder.jpg", "folder.jpg",
		"Cover.jpg", "Album.jpg", "Albumart.jpg", ".Folder.jpg", "Folder.jpg"};

	GFile *file = g_file_new_for_uri (location);
	GFile *parent = g_file_get_parent (file);
	gchar *parent_path = g_file_get_path (parent);
	g_object_unref (file);
	g_object_unref (parent);

	int i;
	gchar *filename;
	GdkPixbuf *pixbuf=NULL;
	for(i=0;i<G_N_ELEMENTS(cover_names);i++)
	{
		filename = g_strconcat (parent_path,"/",cover_names[i],NULL);
		if(g_file_test(filename,G_FILE_TEST_IS_REGULAR|G_FILE_TEST_EXISTS))
		{
			pixbuf = gdk_pixbuf_new_from_file_at_size (filename,200,200,NULL);
			if(pixbuf)
			{
				g_free(filename);
				break;
			}
		}
		g_free(filename);
	}
	
	g_free(parent_path);

	return pixbuf;
}

typedef struct _IdleData
{
	RBArtyPlugin *plugin;
	gchar *album_name;
}IdleData;

static gboolean
rb_arty_cover_fetcher_idle_func(gpointer data)
{
	IdleData *idle_data=data;

	static gint gotmany=0;

	g_print("Album %d *%s* is processing\n",++gotmany,idle_data->album_name);

	//get entry location
	const gchar *location = rb_arty_get_album_first_entry_location(idle_data->plugin,
	                                                     idle_data->album_name);
	//if no entry for album, we are done
	if(location)
	{
		GdkPixbuf *pixbuf = NULL;	

		//find art in folder
		pixbuf = rb_arty_get_album_art_from_location(location);

		if(!pixbuf)
		{
			//try by embeded art and return 
			rb_arty_get_embedded_cover (idle_data->plugin,
			                            idle_data->album_name,
			                            location);
			g_free(idle_data->album_name);
			g_free(idle_data);
			return FALSE;
		}
		else
		{
			//g_print("From file : %s\n",idle_data->album_name);
			rb_arty_save_coverart (idle_data->plugin,
			                            idle_data->album_name,pixbuf);
			g_object_unref (pixbuf);
		}
	}


	//this album is done, looup for next
	idle_data->plugin->priv->cover_fetcher->idle_running = FALSE;
	rb_arty_cover_fetcher_fetch_next(idle_data->plugin->priv->cover_fetcher);

	g_free(idle_data->album_name);
	g_free(idle_data);
	return FALSE;
}

void
rb_arty_cover_fetcher_fetch_next(RbArtyCoverFetcher *fetcher)
{
	if(fetcher->idle_running)
		return;

	//see if iterator is valid
	if(fetcher->list_iter)
	{
		gchar *album_name = fetcher->list_iter->data;
		
		//add idle fetching for given album
		IdleData *data = g_new0(IdleData,1);
		data->plugin = fetcher->plugin;
		data->album_name = g_strdup(album_name);
		
		fetcher->idle_id = g_idle_add (rb_arty_cover_fetcher_idle_func,data);
		fetcher->idle_running = TRUE;

		//go to next
		fetcher->list_iter = fetcher->list_iter->next;

		//update progress bar
		rb_arty_cover_progress_step(fetcher->plugin);
	}
	else
	{
		fetcher->idle_id = 0;
		rb_arty_cover_progress_finish(fetcher->plugin);
	}
}

void 
rb_arty_cover_fetcher_set_albums(RbArtyCoverFetcher *fetcher, GList *list)
{
	//free previous list
	if(fetcher->album_list)
	{
		g_list_foreach(fetcher->album_list,(GFunc)g_free,NULL);
		g_list_free (fetcher->album_list);
		fetcher->album_list=NULL;
		fetcher->list_iter=NULL;
	}

	fetcher->album_list = list;
	fetcher->list_iter = fetcher->album_list;
}

RbArtyCoverFetcher*
rb_arty_cover_fetcher_new(RBArtyPlugin *plugin)
{
	RbArtyCoverFetcher *fetcher;
	fetcher = g_new0(RbArtyCoverFetcher,1);

	fetcher->plugin = plugin;
	fetcher->idle_running = FALSE;
	
	fetcher->album_list=NULL;
	fetcher->list_iter=NULL;
	
	return fetcher;
}

void 
rb_arty_cover_fetcher_free(RbArtyCoverFetcher *fetcher)
{
	if(fetcher->cover)
	{
		//gst_element_set_state (fetcher->cover->pipeline,GST_STATE_NULL);
		g_object_unref(fetcher->cover->pipeline);	
		g_free(fetcher->cover);
		fetcher->cover = NULL;
	}
		
	//free previous list
	if(fetcher->album_list)
	{
		g_list_foreach(fetcher->album_list,(GFunc)g_free,NULL);
		g_list_free (fetcher->album_list);
		fetcher->album_list=NULL;
		fetcher->list_iter=NULL;
	}
	
	g_free (fetcher);
}

//################### fetching progress widgets and operation #################

gboolean 
rb_arty_coverart_exists(RBArtyPlugin *plugin, const gchar *album_name)
{
	gboolean exists = FALSE;
	gchar *escaped = g_uri_escape_string(album_name,NULL,FALSE);
	gchar *filename=g_strconcat(plugin->priv->cache_dir,"/",
	                            escaped,".jpg",NULL);

	exists = g_file_test (filename,G_FILE_TEST_IS_REGULAR|G_FILE_TEST_EXISTS);

	g_free(escaped);
	g_free(filename);

	return exists;
}

void
rb_arty_get_covers_clicked (GtkAction *action, RBArtyPlugin *plugin)
{
	//we are alredy fetching covers
	if(plugin->priv->cover_fetcher->idle_running)
		return;
	
	plugin->priv->nalbums = 0;

	gchar *album_name; 
	GList *albums=NULL;
	GtkTreeIter iter;
	gboolean got;
	
	got = gtk_tree_model_get_iter_from_string (plugin->priv->album_model,&iter,"1");

	while(got)
	{
		//let gui to be responsive
		while (g_main_context_iteration(NULL, FALSE));
		
		gtk_tree_model_get (plugin->priv->album_model,&iter,
		                    RHYTHMDB_PROPERTY_MODEL_COLUMN_TITLE,
		                    &album_name,-1);

		//see if cover already exists
		if(!rb_arty_coverart_exists(plugin,album_name))
		{
			albums = g_list_prepend(albums,album_name);
			plugin->priv->nalbums++;
		}
		else
		{
			g_free(album_name);
		}
		                  
		got = gtk_tree_model_iter_next (plugin->priv->album_model,&iter);
	}

	if(albums)
	{
		albums = g_list_reverse(albums);
		rb_arty_cover_fetcher_set_albums(plugin->priv->cover_fetcher,albums);
		rb_arty_cover_fetcher_fetch_next(plugin->priv->cover_fetcher);

		//show progress bars
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(plugin->priv->fetching_progress)
		                                ,0.0);
		gtk_progress_bar_set_text (GTK_PROGRESS_BAR(plugin->priv->fetching_progress),"");
		gtk_box_pack_start (GTK_BOX(plugin->priv->main_widget),
	                    plugin->priv->fetching_box,FALSE,FALSE,2);
		gtk_widget_show_all (plugin->priv->fetching_box);
		plugin->priv->nth_cover_fetching = 0;
	}
}

static void 
rb_arty_cover_progress_step(RBArtyPlugin *plugin)
{
	plugin->priv->nth_cover_fetching++;

	gdouble f = (gdouble)plugin->priv->nth_cover_fetching/(gdouble)plugin->priv->nalbums;
	gchar *text = g_strdup_printf ("%s %d %s %d %s",_("Fetching coverart"),
	                               plugin->priv->nth_cover_fetching,
	                               _("of"),
	                               plugin->priv->nalbums,
	                               _("albums")
	                               );
	
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(plugin->priv->fetching_progress),f);
	gtk_progress_bar_set_text (GTK_PROGRESS_BAR(plugin->priv->fetching_progress),text);
	
	g_free(text);
}

static void 
rb_arty_cover_progress_finish(RBArtyPlugin *plugin)
{
		gtk_container_remove (GTK_CONTAINER(plugin->priv->main_widget),
	                    plugin->priv->fetching_box);
		gtk_widget_hide_all (plugin->priv->fetching_box);
}

void
rb_arty_cover_fetching_cancel(RBArtyPlugin *plugin)
{
	if(plugin->priv->cover_fetcher->idle_id)
	{
		rb_arty_cover_progress_finish(plugin);
		g_source_remove (plugin->priv->cover_fetcher->idle_id);
		plugin->priv->cover_fetcher->idle_id = 0;

		if(plugin->priv->cover_fetcher->cover &&
		   GST_IS_ELEMENT(plugin->priv->cover_fetcher->cover->src))
		{
			gst_element_set_state (plugin->priv->cover_fetcher->cover->pipeline,GST_STATE_NULL);
			arty_cover_free(plugin->priv->cover_fetcher->cover);
		}
	}
	
	plugin->priv->cover_fetcher->list_iter = NULL;
	plugin->priv->cover_fetcher->idle_running = FALSE;
}

